# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from enum import Enum


class ConfigKeys(Enum):
    KEY_URL = 'url'                 # base url for the Secrets Manager service
    KEY_CLIENT_ID = 'clientId'
    KEY_CLIENT_KEY = 'clientKey'    # The key that is used to identify the client before public key. This is token.
    KEY_APP_KEY = 'appKey'          # The application key with which all secrets are encrypted
    KEY_OWNER_PUBLIC_KEY = 'appOwnerPublicKey'  # The application owner public key, to create records
    KEY_PRIVATE_KEY = 'privateKey'  # The client's private key
    KEY_SERVER_PUBLIC_KEY_ID = 'serverPublicKeyId'  # Which public key should be using?

    KEY_BINDING_TOKEN = 'bat'
    KEY_BINDING_KEY = 'bindingKey'
    KEY_HOSTNAME = 'hostname'

    @classmethod
    def get_enum(cls, value):
        for e in cls:
            # Check if the value passed is the value of the enum key, or the enum key itself.
            if e.value == value or e == value:
                return e
        return None



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import hmac
import logging
import os
import sys
from distutils.util import strtobool
import re
import json
from http import HTTPStatus

import requests

from keeper_secrets_manager_core import utils, helpers
from keeper_secrets_manager_core.configkeys import ConfigKeys
from keeper_secrets_manager_core.crypto import CryptoUtils
from keeper_secrets_manager_core.dto.dtos import Folder, Record, RecordCreate, SecretsManagerResponse, AppData, \
    KeeperFileUpload
from keeper_secrets_manager_core.dto.payload import GetPayload, UpdatePayload, TransmissionKey, \
    EncryptedPayload, KSMHttpResponse, CreatePayload, FileUploadPayload, DeletePayload
from keeper_secrets_manager_core.exceptions import KeeperError
from keeper_secrets_manager_core.keeper_globals import keeper_secrets_manager_sdk_client_id, keeper_public_keys, \
    logger_name, keeper_servers
from keeper_secrets_manager_core.storage import FileKeyValueStorage, KeyValueStorage, InMemoryKeyValueStorage
from keeper_secrets_manager_core.utils import base64_to_bytes, dict_to_json, \
    url_safe_str_to_bytes, bytes_to_base64, generate_random_bytes, now_milliseconds, string_to_bytes, json_to_dict, \
    bytes_to_string


def find_secrets_by_title(record_title, records):
    # Find all records with specified title
    records = records or []
    return [x for x in records if x.title == record_title]

def find_secret_by_title(record_title, records):
    # Find first record with specified title
    records = records or []
    return next((x for x in records if x.title == record_title), None)


class SecretsManager:

    notation_prefix = "keeper"
    default_key_id = "10"

    # Field types that can be inflated. Used for notation.
    inflate_ref_types = {
        "addressRef": ["address"],
        "cardRef": ["paymentCard", "text", "pinCode", "addressRef"]
    }

    def __init__(self,
                 token=None, hostname=None, verify_ssl_certs=True, config=None, log_level=None,
                 custom_post_function=None):

        # Make sure the Python is 3.6 or higher. We'll handle Python 4 in the future :)
        python_version = sys.version_info
        if python_version.major < 3 or (python_version.major == 3 and python_version.minor < 6):
            raise Exception("KSM SDK requires Python 3.6 or greater")

        self.token = None
        self.hostname = None

        # If the config is not defined and the KSM_CONFIG env var exists, get the config from the env var.
        if config is None and os.environ.get("KSM_CONFIG") is not None:
            config = InMemoryKeyValueStorage(os.environ.get("KSM_CONFIG"))
        elif token:

            token = token.strip()
            token_parts = token.split(":")

            if len(token_parts) == 1:
                if not hostname:
                    raise ValueError('The hostname must be present in the token or provided as a parameter')

                self.token = token
                self.hostname = hostname
            else:
                token_host = keeper_servers.get(token_parts[0].upper())

                if token_host:
                    # meaning that token contained abbreviation:
                    #   ex. 'US:ONE_TIME_TOKEN'
                    self.hostname = token_host
                else:
                    # meaning that token contained url prefix:
                    #   ex. ksm.company.com:ONE_TIME_TOKEN
                    self.hostname = token_parts[0]

                self.token = token_parts[1]

        # Init the log, create a logger for the core.
        self._init_logger(log_level=log_level)
        self.logger = logging.getLogger(logger_name)

        # Accept the env var KSM_SKIP_VERIFY. Modules like 'requests' already use it.
        self.verify_ssl_certs = verify_ssl_certs
        if os.environ.get("KSM_SKIP_VERIFY") is not None:
            # We need to flip the value of KSM_SKIP_VERIFY, if true, we want verify_ssl_certs to be false.
            self.verify_ssl_certs = not bool(strtobool(os.environ.get("KSM_SKIP_VERIFY")))

        self.custom_post_function = custom_post_function

        if config is None:
            config = FileKeyValueStorage()

        # If the server or client key are set in the args, make sure they makes it's way into the config. They
        # will override what is already in the config if they exist.
        if self.token is not None:
            config.set(ConfigKeys.KEY_CLIENT_KEY, self.token)
        if self.hostname is not None:
            config.set(ConfigKeys.KEY_HOSTNAME, self.hostname)

        # Make sure our public key id is set and pointing an existing key.
        if config.get(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID) is None:
            self.logger.debug("Setting public key id to the default: {}".format(SecretsManager.default_key_id))
            config.set(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID, SecretsManager.default_key_id)
        elif config.get(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID) not in keeper_public_keys:
            self.logger.debug("Public key id {} does not exists, set to default : {}".format(
                config.get(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID),
                SecretsManager.default_key_id))
            config.set(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID, SecretsManager.default_key_id)

        self.config: KeyValueStorage = config

        self._init()

    @staticmethod
    def _init_logger(log_level=None):

        logger = logging.getLogger(logger_name)

        # If the log level was passed in, then we want to set up the logging. If not, there is no logging.
        if log_level is not None:

            # If the log level is a string, get the enum value.
            if type(log_level) is str:
                valid_log_levels = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"]
                if log_level not in valid_log_levels:
                    raise ValueError("Log level {} is invalid. Valid values are: {}".format(
                        log_level, ", ".join(valid_log_levels)))
                log_level = getattr(logging, log_level)

            logger.disabled = False
            logger.setLevel(log_level)

            # If we don't have my handlers, set the logger up with default settings. The handler and formatters
            # can be set outside of the SDK if an integration or company
            if len(logger.handlers) == 0:
                # Default to stderr
                sh = logging.StreamHandler()
                sh.setLevel(log_level)

                sh.setFormatter(logging.Formatter("%(asctime)s | %(name)s | %(levelname)s | %(message)s"))
                logger.addHandler(sh)
        else:
            logger.disabled = True

    def _init(self):

        client_id = self.config.get(ConfigKeys.KEY_CLIENT_ID)

        if client_id:
            self.logger.debug("Already bound")

            if self.config.get(ConfigKeys.KEY_CLIENT_KEY):
                self.config.delete(ConfigKeys.KEY_CLIENT_KEY)

        else:

            existing_secret_key = self.load_secret_key()

            if existing_secret_key is None:
                raise ValueError("Cannot locate One Time Token.")

            existing_secret_key_bytes = url_safe_str_to_bytes(existing_secret_key)
            digest = 'sha512'
            existing_secret_key_hash = bytes_to_base64(hmac.new(existing_secret_key_bytes,
                                                                      b'KEEPER_SECRETS_MANAGER_CLIENT_ID',
                                                                      digest).digest())

            self.config.delete(ConfigKeys.KEY_CLIENT_ID)
            self.config.delete(ConfigKeys.KEY_PRIVATE_KEY)
            if self.config.get(ConfigKeys.KEY_CLIENT_ID):
                self.config.delete(ConfigKeys.KEY_APP_KEY)

            self.config.set(ConfigKeys.KEY_CLIENT_ID, existing_secret_key_hash)

            private_key_str = self.config.get(ConfigKeys.KEY_PRIVATE_KEY)

            if not private_key_str:
                private_key_der = CryptoUtils.generate_private_key_der()
                self.config.set(ConfigKeys.KEY_PRIVATE_KEY, bytes_to_base64(private_key_der))

        if not self.verify_ssl_certs:
            self.logger.warning("WARNING: Running without SSL cert verification. "
                                "Execute 'SecretsManager(..., verify_ssl_certs=True)' or 'KSM_SKIP_VERIFY=FALSE' "
                                "to enable verification.")

    def load_secret_key(self):

        """Returns client_id from the environment variable, config file, or in the code"""

        # Case 1: Environment Variable
        env_secret_key = os.getenv('KSM_TOKEN')

        current_secret_key = None

        if env_secret_key:
            current_secret_key = env_secret_key
            self.logger.info("Secret key found in environment variable")

        # Case 2: Code
        if not current_secret_key:
            code_secret_key = self.token

            if code_secret_key:
                current_secret_key = code_secret_key
                self.logger.info("Secret key found in code")

        # Case 3: Config storage
        if not current_secret_key:
            config_secret_key = self.config.get(ConfigKeys.KEY_CLIENT_KEY)

            if config_secret_key:
                current_secret_key = config_secret_key
                self.logger.info("Secret key found in configuration file")

        return current_secret_key

    @staticmethod
    def generate_transmission_key(key_id):
        transmission_key = utils.generate_random_bytes(32)

        if key_id not in keeper_public_keys:
            ValueError("The public key id {} does not exist.".format(key_id))

        server_public_raw_key_bytes = url_safe_str_to_bytes(keeper_public_keys[key_id])

        encrypted_key = CryptoUtils.public_encrypt(transmission_key, server_public_raw_key_bytes)

        return TransmissionKey(key_id, transmission_key, encrypted_key)

    @staticmethod
    def encrypt_and_sign_payload(storage, transmission_key, payload):

        if not (
                isinstance(payload, GetPayload) or
                isinstance(payload, UpdatePayload) or
                isinstance(payload, CreatePayload) or
                isinstance(payload, FileUploadPayload) or
                isinstance(payload, DeletePayload)):
            raise Exception('Unknown payload type "%s"' % payload.__class__.__name__)

        payload_json_str = dict_to_json(payload.__dict__)
        payload_bytes = utils.string_to_bytes(payload_json_str)

        encrypted_payload = CryptoUtils.encrypt_aes(payload_bytes, transmission_key.key)

        encrypted_key = transmission_key.encryptedKey
        signature_base = encrypted_key + encrypted_payload

        private_key = storage.get(ConfigKeys.KEY_PRIVATE_KEY)
        pk = CryptoUtils.der_base64_private_key_to_private_key(private_key)
        signature = CryptoUtils.sign(signature_base, pk)

        return EncryptedPayload(
            encrypted_payload,
            signature
        )

    @staticmethod
    def prepare_delete_payload(storage, record_uids):
        payload = DeletePayload()

        payload.clientVersion = keeper_secrets_manager_sdk_client_id
        payload.clientId = storage.get(ConfigKeys.KEY_CLIENT_ID)
        payload.recordUids = record_uids

        return payload

    @staticmethod
    def prepare_get_payload(storage, records_filter):

        payload = GetPayload()

        payload.clientVersion = keeper_secrets_manager_sdk_client_id
        payload.clientId = storage.get(ConfigKeys.KEY_CLIENT_ID)

        app_key_str = storage.get(ConfigKeys.KEY_APP_KEY)

        if not app_key_str:

            public_key_bytes = CryptoUtils.extract_public_key_bytes(storage.get(ConfigKeys.KEY_PRIVATE_KEY))
            public_key_base64 = bytes_to_base64(public_key_bytes)
            # passed once when binding
            payload.publicKey = public_key_base64

        if records_filter:
            payload.requestedRecords = records_filter

        return payload

    @staticmethod
    def prepare_create_payload(storage, folder_uid, record_data_json_str, folder_key):

        owner_public_key = storage.get(ConfigKeys.KEY_OWNER_PUBLIC_KEY)

        if not owner_public_key:
            raise KeeperError('Unable to create record - owner key is missing. Looks like application was created '
                              'using out date client (Web Vault or Commander)')

        owner_public_key_bytes = url_safe_str_to_bytes(owner_public_key)

        if not folder_key:
            raise KeeperError('Unable to create record - folder key for ' + folder_uid + ' is missing')

        record_bytes = ""
        record_key = generate_random_bytes(32)
        record_uid = generate_random_bytes(16)

        record_data_bytes = utils.string_to_bytes(record_data_json_str)
        record_data_encrypted = CryptoUtils.encrypt_aes(record_data_bytes, record_key)

        record_key_encrypted = CryptoUtils.public_encrypt(record_key, owner_public_key_bytes)

        folder_key_encrypted = CryptoUtils.encrypt_aes(record_key, folder_key)

        payload = CreatePayload()

        payload.clientVersion = keeper_secrets_manager_sdk_client_id
        payload.clientId = storage.get(ConfigKeys.KEY_CLIENT_ID)
        payload.recordUid = CryptoUtils.bytes_to_url_safe_str(record_uid)
        payload.recordKey = bytes_to_base64(record_key_encrypted)
        payload.folderUid = folder_uid
        payload.folderKey = bytes_to_base64(folder_key_encrypted)
        payload.data = bytes_to_base64(record_data_encrypted)

        return payload

    @staticmethod
    def prepare_update_payload(storage, record):

        payload = UpdatePayload()

        payload.clientVersion = keeper_secrets_manager_sdk_client_id
        payload.clientId = storage.get(ConfigKeys.KEY_CLIENT_ID)

        # for update, uid of the record
        payload.recordUid = record.uid
        payload.revision = record.revision

        raw_json_bytes = utils.string_to_bytes(record.raw_json)
        encrypted_raw_json_bytes = CryptoUtils.encrypt_aes(raw_json_bytes, record.record_key_bytes)

        payload.data = bytes_to_base64(encrypted_raw_json_bytes)

        return payload

    @staticmethod
    def prepare_file_upload_payload(storage, owner_record, file: KeeperFileUpload):

        owner_public_key = storage.get(ConfigKeys.KEY_OWNER_PUBLIC_KEY)

        if not owner_public_key:
            raise KeeperError('Unable to upload file - owner key is missing. Looks like application was created '
                              'using out date client (Web Vault or Commander)')

        owner_public_key_bytes = url_safe_str_to_bytes(owner_public_key)

        file_record_dict = {
            'name': file.Name,
            'size': len(file.Data),
            'title': file.Title,
            'lastModified': now_milliseconds(),
            'type': file.Type
        }

        file_record_json_str = dict_to_json(file_record_dict)
        file_record_bytes = utils.string_to_bytes(file_record_json_str)

        file_record_key = generate_random_bytes(32)
        file_record_uid = generate_random_bytes(16)
        file_record_uid_str = CryptoUtils.bytes_to_url_safe_str(file_record_uid)

        encrypted_file_record_bytes = CryptoUtils.encrypt_aes(file_record_bytes, file_record_key)
        encrypted_file_record_key = CryptoUtils.public_encrypt(file_record_key, owner_public_key_bytes)
        encrypted_link_key_bytes = CryptoUtils.encrypt_aes(file_record_key, owner_record.record_key_bytes)

        encrypted_file_data = CryptoUtils.encrypt_aes(file.Data, file_record_key)

        # Add fileRef here
        rec_dict = owner_record.dict
        fields = rec_dict.get('fields')

        file_refs = [f for f in fields if f['type'] == 'fileRef']

        if not file_refs:
            fields.append({'type': 'fileRef', 'value': [file_record_uid_str]})
        else:
            file_uid_list = file_refs[0].get('value')
            file_uid_list.append(file_record_uid_str)

        owner_record.raw_json = utils.dict_to_json(rec_dict)

        owner_record_bytes = string_to_bytes(owner_record.raw_json)

        encrypted_owner_record_bytes = CryptoUtils.encrypt_aes(owner_record_bytes, owner_record.record_key_bytes)
        encrypted_owner_record_str = CryptoUtils.bytes_to_url_safe_str(encrypted_owner_record_bytes)

        payload = FileUploadPayload()
        payload.clientVersion = keeper_secrets_manager_sdk_client_id
        payload.clientId = storage.get(ConfigKeys.KEY_CLIENT_ID)
        payload.fileRecordUid = file_record_uid_str
        payload.fileRecordData = CryptoUtils.bytes_to_url_safe_str(encrypted_file_record_bytes)
        payload.fileRecordKey = bytes_to_base64(encrypted_file_record_key)
        payload.ownerRecordUid = owner_record.uid

        payload.ownerRecordData = encrypted_owner_record_str

        payload.linkKey = bytes_to_base64(encrypted_link_key_bytes)

        payload.fileSize = len(encrypted_file_data)
        return {
            'payload': payload,
            'encryptedFileData': encrypted_file_data
        }

    def _post_query(self, path, payload):

        keeper_server = helpers.get_server(self.hostname, self.config)
        url = "https://%s/api/rest/sm/v1/%s" % (keeper_server, path)

        while True:

            transmission_key_id = self.config.get(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID)
            transmission_key = self.generate_transmission_key(transmission_key_id)
            encrypted_payload_and_signature = self.encrypt_and_sign_payload(self.config, transmission_key, payload)

            if self.custom_post_function and path == 'get_secret':
                ksm_rs = self.custom_post_function(url, transmission_key, encrypted_payload_and_signature, self.verify_ssl_certs)
            else:
                ksm_rs = self.post_function(url, transmission_key, encrypted_payload_and_signature, self.verify_ssl_certs)

            # If we are ok, then break out of the while loop
            if ksm_rs.status_code == 200:
                break

            # Handle the error. Handling will throw an exception if it doesn't want us to retry.
            self.handler_http_error(ksm_rs.http_response)

        if ksm_rs.data:
            return CryptoUtils.decrypt_aes(ksm_rs.data, transmission_key.key)
        else:
            return ksm_rs.data

    @staticmethod
    def post_function(url, transmission_key, encrypted_payload_and_signature, verify_ssl_certs=True):

        request_headers = {
            'Content-Type': 'application/octet-stream',
            'Content-Length': str(len(encrypted_payload_and_signature.encrypted_payload)),
            'PublicKeyId': str(transmission_key.publicKeyId),
            'TransmissionKey': bytes_to_base64(transmission_key.encryptedKey),
            'Authorization': 'Signature %s' % bytes_to_base64(encrypted_payload_and_signature.signature)
        }

        rs = requests.post(
            url,
            headers=request_headers,
            data=encrypted_payload_and_signature.encrypted_payload,
            verify=verify_ssl_certs
        )

        ksm_rs = KSMHttpResponse(rs.status_code, rs.content, rs)

        return ksm_rs

    @staticmethod
    def __upload_file_function(url, upload_parameters, encrypted_file_data):
        """Upload file to the server"""
        files = {'file': encrypted_file_data}

        rs = requests.post(url,
                           data=upload_parameters,
                           files=files,
                           )

        rs_status_code = rs.status_code
        rs_data = rs.text

        return {
            'isOk': rs.ok,
            'statusCode': rs_status_code,
            'data': rs_data
        }

    def handler_http_error(self, rs):

        log_level = logging.ERROR
        try:
            # Decode the JSON content, throw exception if not JSON
            response_dict = utils.json_to_dict(rs.text)
            if response_dict is None:
                raise json.JSONDecodeError("Not JSON", "NONE", 0)

            # Try to get the error from result_code, then from error.
            error = response_dict.get('result_code', response_dict.get('error'))

            if error == 'invalid_client_version':
                self.logger.error("Client version {} was not registered in the backend".format(
                    keeper_secrets_manager_sdk_client_id))
                msg = response_dict.get('additional_info')

            # The server wants us to use a different public key.
            elif error == 'key':
                key_id = response_dict.get("key_id")
                self.logger.info("Server has requested we use public key {}".format(key_id))

                if key_id is None:
                    raise ValueError("The public key is blank from the server")
                elif str(key_id) not in keeper_public_keys:
                    raise ValueError("The public key at {} does not exist in the SDK".format(key_id))

                self.config.set(ConfigKeys.KEY_SERVER_PUBLIC_KEY_ID, str(key_id))

                # This is not an error, it's info. Make it so the 'finally' display info about the
                # key change.
                log_level = logging.DEBUG

                # The only non-exception exit from this method
                return True
            else:
                msg = "Error: {}, message={}".format(error, response_dict.get('message', "NA"))

            raise KeeperError(msg)

        except json.JSONDecodeError as _:
            # The content wasn't JSON. Let the catch-all exception at the end handle it.
            pass
        except KeeperError as err:
            # This was one of our exceptions, just rethrow it.
            raise err
        finally:
            self.logger.log(log_level, "Error: {} (http error code {}): {}".format(
                rs.reason,
                rs.status_code,
                rs.text
            ))

        # This is a unknown error, not one of ours, just throw a HTTPError
        raise requests.HTTPError(rs.text)

    def fetch_and_decrypt_secrets(self, record_filter=None):

        payload = SecretsManager.prepare_get_payload(self.config, records_filter=record_filter)

        decrypted_response_bytes = self._post_query(
            'get_secret',
            payload
        )

        decrypted_response_str = utils.bytes_to_string(decrypted_response_bytes)
        decrypted_response_dict = utils.json_to_dict(decrypted_response_str)

        records = []
        shared_folders = []

        just_bound = False

        if decrypted_response_dict.get('encryptedAppKey'):
            just_bound = True

            encrypted_master_key = url_safe_str_to_bytes(decrypted_response_dict.get('encryptedAppKey'))
            client_key = url_safe_str_to_bytes(self.config.get(ConfigKeys.KEY_CLIENT_KEY))
            secret_key = CryptoUtils.decrypt_aes(encrypted_master_key, client_key)
            self.config.set(ConfigKeys.KEY_APP_KEY, bytes_to_base64(secret_key))

            self.config.delete(ConfigKeys.KEY_CLIENT_KEY)

            if decrypted_response_dict.get('appOwnerPublicKey'):
                appOwnerPublicKeyBytes = url_safe_str_to_bytes(decrypted_response_dict.get('appOwnerPublicKey'))
                self.config.set(ConfigKeys.KEY_OWNER_PUBLIC_KEY, bytes_to_base64(appOwnerPublicKeyBytes))

        else:
            secret_key = base64_to_bytes(self.config.get(ConfigKeys.KEY_APP_KEY))

        records_resp = decrypted_response_dict.get('records')
        folders_resp = decrypted_response_dict.get('folders')

        self.logger.debug("Individual record count: {}".format(len(records_resp or [])))
        self.logger.debug("Folder count: {}".format(len(folders_resp or [])))

        if records_resp:
            for r in records_resp:
                record = Record(r, secret_key)
                records.append(record)

        if folders_resp:
            for f in folders_resp:
                folder = Folder(f, secret_key)
                records.extend(folder.records)
                shared_folders.append(folder)

        self.logger.debug("Total record count: {}".format(len(records)))

        sm_response = SecretsManagerResponse()

        if 'appData' in decrypted_response_dict:
            app_data_json = CryptoUtils.decrypt_aes(
                url_safe_str_to_bytes(decrypted_response_dict['appData']),
                base64_to_bytes(self.config.get(ConfigKeys.KEY_APP_KEY))
            )

            app_data_dict = utils.json_to_dict(app_data_json)

            sm_response.appData = AppData(title=app_data_dict['title'], app_type=app_data_dict['type'])
        else:
            sm_response.appData = AppData()

        if 'expiresOn' in decrypted_response_dict:
            sm_response.expiresOn = decrypted_response_dict.get('expiresOn')

        if 'warnings' in decrypted_response_dict:
            sm_response.warnings = decrypted_response_dict.get('warnings')

        sm_response.records = records
        sm_response.folders = shared_folders
        sm_response.justBound = just_bound

        return sm_response

    def get_secrets(self, uids=None, full_response=False):
        """
        Retrieve all records associated with the given application
        """

        if isinstance(uids, str):
            uids = [uids]

        records_resp = self.fetch_and_decrypt_secrets(uids)

        if records_resp.justBound:
            records_resp = self.fetch_and_decrypt_secrets(uids)

        # Log warnings we got from the server
        # Will only be displayed if logging is enabled:
        if records_resp.warnings:
            for warning in records_resp.warnings:
                self.logger.warning(warning)

        if full_response:
            return records_resp
        else:
            records = records_resp.records or []

            return records

    def get_secrets_by_title(self, record_title):
        """
        Retrieve all secret records with specified title
        """

        recs = self.get_secrets()
        records = [x for x in recs if x.title == record_title]

        return records

    def get_secret_by_title(self, record_title):
        """
        Retrieve first secret record with specified title
        """

        records = self.get_secrets_by_title(record_title) or []
        return next((iter(records)), None)

    def delete_secret(self, record_uids):
        """
        Delete secret records with specified uids
        """

        if isinstance(record_uids, str):
            record_uids = [record_uids]

        payload = SecretsManager.prepare_delete_payload(self.config, record_uids=record_uids)
        response = self._post_query('delete_secret', payload)
        response_str = bytes_to_string(response)
        response_dict = json_to_dict(response_str)

        return response_dict.get('records')

    def create_secret(self, folder_uid, record_data):

        #   Backend only need a JSON string of the record, so we have different ways of handing data:
        #       - providing data as JSON string
        #       - providing data as dictionary
        #       - providing data as CreateRecord object
        #
        #   For now we will only allow CreateRecord objects

        record_data_json_str = None

        if isinstance(record_data, RecordCreate):
            record_data_json_str = record_data.to_json()
        else:
            raise KeeperError('New record data has to be a valid ' + RecordCreate.__name__ + ' object')

        # if isinstance(record_data, RecordV3):
        #     record_data_json_str = record_data.to_json()
        # elif isinstance(record_data, dict):
        #     record_data_json_str = dict_to_json(record_data)
        # elif isinstance(record_data, str):
        #     if not is_json(record_data):
        #         raise KeeperError('Record data has to be a valid JSON string.')
        #
        #     record_data_json_str = record_data

        # Since we don't know folder's key where this record will be
        # placed in, currently we have to retrieve all data that is share to
        # this device/client and look for the folder's key in the returned
        # folder data

        records_and_folders_response = self.get_secrets(full_response=True)

        found_folder = helpers.get_folder_key(folder_uid=folder_uid, secrets_and_folders=records_and_folders_response)

        if not found_folder:
            raise KeeperError('Folder uid=' + folder_uid + ' was not retrieved. If you are creating a record to a '
                              'folder folder that you know exists, make sure that at least one record is present in '
                              'the prior to adding a record to the folder.')

        payload = SecretsManager.prepare_create_payload(self.config, folder_uid, record_data_json_str, found_folder.key)

        self._post_query('create_secret', payload)

        return payload.recordUid

    def upload_file(self, owner_record, file: KeeperFileUpload):
        """
        Upload file using provided file upload object
        """

        self.logger.info(f"Uploading file: {file.Name} to record uid {owner_record.uid}")

        self.logger.debug(f"Preparing upload payload. owner_record.uid=[{owner_record.uid}], file name: [{file.Name}], file size: [{len(file.Data)}]")

        upload_payload = self.prepare_file_upload_payload(self.config, owner_record, file)
        payload = upload_payload.get('payload')
        encrypted_file_data = upload_payload.get('encryptedFileData')

        self.logger.debug(f"Posting prepare data")
        response_data = self._post_query('add_file', payload)

        response_json_str = bytes_to_string(response_data)
        response_dict = json_to_dict(response_json_str)
        upload_url = response_dict.get('url')
        parameters_json_str = response_dict.get('parameters')
        parameters_dict = json_to_dict(parameters_json_str)

        self.logger.debug(f"Uploading file data: upload url=[{upload_url}], file name: [{file.Name}], encrypted file size: [{len(encrypted_file_data)}]")
        upload_result = SecretsManager.__upload_file_function(upload_url, parameters_dict, encrypted_file_data)

        self.logger.debug(f"Finished uploading file data. Status code: {upload_result.get('statusCode')}, response data: {upload_result.get('data')}")

        if not upload_result.get('isOk'):
            raise KeeperError('Failed to upload a file')
        else:
            return payload.fileRecordUid

    def upload_file_path(self, owner_record, file_path):
        """
        Upload file using provided file path
        """

        file_to_upload = KeeperFileUpload.from_file(file_path)

        return self.upload_file(owner_record, file_to_upload)

    def save(self, record):
        """
        Save updated secret values
        """

        self.logger.info("Updating record uid: %s" % record.uid)

        payload = SecretsManager.prepare_update_payload(self.config, record)

        self._post_query(
            'update_secret',
            payload
        )

        return True

    def get_notation(self, url):

        """Simple string notation to get a value

        * A system of figures or symbols used in a specialized field to represent numbers, quantities, tones,
          or values.

        <uid>/<field|custom_field|file>/<label|type>[INDEX][FIELD]

        Example:

            RECORD_UID/field/password                => MyPassword
            RECORD_UID/field/password[0]             => MyPassword
            RECORD_UID/field/password[]              => ["MyPassword"]
            RECORD_UID/custom_field/name[first]      => John
            RECORD_UID/custom_field/name[last]       => Smith
            RECORD_UID/custom_field/phone[0][number] => "555-5555555"
            RECORD_UID/custom_field/phone[1][number] => "777-7777777"
            RECORD_UID/custom_field/phone[]          => [{"number": "555-555...}, { "number": "777.....}]
            RECORD_UID/custom_field/phone[0]         => [{"number": "555-555...}]

        """

        # If the URL starts with keeper:// we want to remove it.
        if url.startswith(SecretsManager.notation_prefix) is True:
            url_parts = url.split('//')
            try:
                url = url_parts[1]
                if url is None:
                    # Get the except below handle it
                    raise ValueError()
            except IndexError:
                raise ValueError("Keeper url missing information about the uid, field type, and field key.")

        try:
            (uid, file_data_type, key) = url.split('/')
        except Exception as _:
            raise ValueError("Could not parse the notation {}. Is it valid?".format(url))

        if uid is None:
            raise ValueError("UID is missing the in the keeper url.")
        if file_data_type is None:
            raise ValueError("file type is missing the in the keeper url.")
        if key is None:
            raise ValueError("file key is missing the in the keeper url.")

        # By default, we want to return a single value, which is the first item in the array
        return_single = True
        index = 0
        dict_key = None

        # Check it see if the key has a predicate, possibly with an index.
        predicate = re.search(r'\[.*]', key)
        if predicate is not None:

            # If we do, get the predicate and remove the brackets, to get the index if one exists
            match = predicate.group()

            predicate_parts = match.split("]")
            while "" in predicate_parts:
                predicate_parts.remove("")

            if len(predicate_parts) == 0:
                raise ValueError("The predicate of the notation appears to be invalid. Syntax error?")
            if len(predicate_parts) > 2:
                raise ValueError("The predicate of the notation appears to be invalid. Too many [], max 2 allowed.")

            # This will remove the preceding '['
            first_predicate = predicate_parts[0][1:]
            if first_predicate is not None:
                # Is the first predicate an index into an array?
                if first_predicate.isdigit() is True:
                    index = int(first_predicate)
                # Is the first predicate a key to a dictionary?
                elif re.match(r'^[a-zA-Z0-9_]+$', first_predicate):
                    dict_key = first_predicate
                # Else it was an array indicator. Return all the values.
                else:
                    return_single = False

            if len(predicate_parts) == 2:
                if return_single is False:
                    raise ValueError("If the second [] is a dictionary key, the first [] needs to have any index.")
                # Remove the preceding '['
                second_predicate = predicate_parts[1][1:]
                if second_predicate.isdigit() is True:
                    raise ValueError("The second [] can only by a key for the dictionary. It cannot be an index.")
                # Is the first predicate a key to a dictionary?
                elif re.match(r'^[a-zA-Z0-9_]+$', second_predicate):
                    dict_key = second_predicate
                else:
                    raise ValueError("The second [] must have key for the dictionary. Cannot be blank.")

            # Remove the predicate from the key, if it exists
            key = re.sub(r'\[.*', '', key)

        records = self.get_secrets([uid])
        if len(records) == 0:
            raise ValueError("Could not find a record with the UID {}".format(uid))

        record = records[0]

        field_type = None
        if file_data_type == "field":
            field = record.get_standard_field(key)
            if field is None:
                raise ValueError("Cannot find standard field {}".format(key))
            value = field.get("value")
            field_type = field.get("type")
        elif file_data_type == "custom_field":
            field = record.get_custom_field(key)
            if field is None:
                raise ValueError("Cannot find custom field {}".format(key))
            value = field.get("value")
            field_type = field.get("type")
        elif file_data_type == "file":
            file = record.find_file_by_title(key)
            if file is None:
                raise FileNotFoundError("Cannot find the file {} in record {}.".format(key, uid))
            value = file.get_file_data()
        else:
            raise ValueError("Field type of {} is not valid.".format(file_data_type))

        # Inflate the value if its part of list of types to inflate. This will request additional records
        # from secrets manager.
        if field_type in SecretsManager.inflate_ref_types:
            value = self.inflate_field_value(value, SecretsManager.inflate_ref_types[field_type])

        ret = value
        if return_single is True and type(value) is list:
            if len(value) == 0:
                return None
            try:
                ret = value[index]
                if dict_key is not None:
                    if dict_key not in ret:
                        raise ValueError("Cannot find the dictionary key {} in the value".format(dict_key))
                    ret = ret[dict_key]
            except IndexError:
                raise ValueError("The value at index {} does not exist for {}.".format(index, url))
        return ret

    @staticmethod
    def get_inflate_ref_types(field_type):
        return SecretsManager.inflate_ref_types.get(field_type, [])

    def inflate_field_value(self, uids, replace_fields):

        # The replacement value
        value = []

        # Get the record and make a lookup for them.
        lookup = {}
        records = self.get_secrets(uids)
        for record in records:
            lookup[record.uid] = record

        for uid in uids:
            record = lookup.get(uid)
            new_value = None
            if record is not None:
                for replacement_key in replace_fields:
                    # Replacement are always in the standard fields.
                    real_field = record.get_standard_field(replacement_key)

                    # If we can't find it, move onto the next type. There might be a problem with the record.
                    if real_field is None:
                        self.logger.debug("Cannot find {} in the fields in record UID {} for inflation. Skipping". \
                                          format(replacement_key, uid))
                        continue
                    real_values = real_field.get("value", [])
                    if len(real_values) > 0:
                        real_value = real_values[0]
                        if real_value is not None:

                            # Do we need to replace a value in our real value?
                            if replacement_key in SecretsManager.inflate_ref_types:
                                real_value = self.inflate_field_value([real_value], SecretsManager.inflate_ref_types[
                                    replacement_key])
                                if real_value is not None:
                                    real_value = real_value[0]

                            # If we don't have value, just use the real value. It might be a str or a dict. We
                            # do this because the value of the real_value might be the final value. So it might
                            # be a str or dict, just leave it that way for now.
                            if new_value is None:
                                new_value = real_value
                            else:
                                # If we need to add a new key/value use the label and fall back the the type
                                label = real_field.get("label", real_field.get("type"))

                                # Since we have more than 1 value, convert the value to a dict if it is not already
                                # one.
                                if type(new_value) is not dict:
                                    new_value = {
                                        label: new_value
                                    }
                                # If the real_value is a dict, then copy the k/v pair into the value
                                if type(real_value) is dict:
                                    for k, v in real_value.items():
                                        new_value[k] = v
                                # Else the real_value is str, use the label/type as a key
                                else:
                                    new_value[label] = real_value
            if new_value is not None:
                value.append(new_value)

        return value


class KSMCache:
    # Allow the directory that will contain the cache to be set with environment variables. If not set, the
    # cache file will be create in the current working directory.
    kms_cache_file_name = os.path.join(os.environ.get("KSM_CACHE_DIR", ""), 'ksm_cache.bin')

    @staticmethod
    def save_cache(data):
        cache_file = open(KSMCache.kms_cache_file_name, 'wb')
        cache_file.write(data)
        cache_file.close()

    @staticmethod
    def get_cached_data():
        cache_file = open(KSMCache.kms_cache_file_name, 'rb')
        cache_data = cache_file.read()
        cache_file.close()
        return cache_data

    @staticmethod
    def remove_cache_file():
        if os.path.exists(KSMCache.kms_cache_file_name) is True:
            os.unlink(KSMCache.kms_cache_file_name)

    @staticmethod
    def caching_post_function(url, transmission_key, encrypted_payload_and_signature, verify_ssl_certs=True):

        try:

            ksm_rs = SecretsManager.post_function(url, transmission_key, encrypted_payload_and_signature, verify_ssl_certs)

            if ksm_rs.status_code == 200:
                KSMCache.save_cache(transmission_key.key + ksm_rs.data)
                return ksm_rs
        except:
            cached_data = KSMCache.get_cached_data()
            cached_transmission_key = cached_data[:32]
            transmission_key.key = cached_transmission_key
            data = cached_data[32:len(cached_data)]

            ksm_rs = KSMHttpResponse(HTTPStatus.OK, data, None)

        return ksm_rs



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import base64
import logging
import os
import json

# An Interface for different storage types
import errno
from json import JSONDecodeError

from keeper_secrets_manager_core import exceptions, utils
from keeper_secrets_manager_core.configkeys import ConfigKeys
from keeper_secrets_manager_core.keeper_globals import logger_name
from keeper_secrets_manager_core.utils import ENCODING, json_to_dict, set_config_mode, check_config_mode


class KeyValueStorage:
    """ Interface for the key value storage"""

    def read_storage(self):
        pass

    def save_storage(self, updated_config):
        pass

    def get(self, key: ConfigKeys):
        pass

    def set(self, key: ConfigKeys, value):
        pass

    def delete(self, key: ConfigKeys):
        pass

    def delete_all(self):
        pass

    def contains(self, key: ConfigKeys):
        pass

    def is_empty(self):
        pass


class FileKeyValueStorage(KeyValueStorage):
    """ File based implementation of the key value storage"""

    default_config_file_location = "client-config.json"

    def __init__(self, config_file_location=None):

        if config_file_location is None:
            config_file_location = os.environ.get("KSM_CONFIG_FILE",
                                                  FileKeyValueStorage.default_config_file_location)

        self.default_config_file_location = config_file_location

    def read_storage(self):

        self.create_config_file_if_missing()

        try:
            check_config_mode(self.default_config_file_location)

            with open(self.default_config_file_location, "r", encoding=ENCODING) as fh:
                config_data = None
                try:
                    config_data = fh.read()
                    config = json.loads(config_data)
                except UnicodeDecodeError:
                    logging.getLogger(logger_name).error("Config file is not utf-8 encoded.")
                    raise Exception("{} is not a utf-8 encoded file".format(self.default_config_file_location))
                except JSONDecodeError as err:

                    # If the JSON file was not empty, it's a legit JSON error. Throw an exception.
                    if config_data is not None and config_data.strip() != "":
                        raise Exception("{} may contain JSON format problems or is not utf-8 encoded"
                                        ": {}".format(self.default_config_file_location, err))

                    # If it was an empty file, overwrite with the JSON config
                    logging.getLogger(logger_name).warning("Looks like config file is empty.")
                    config = {}
                    self.save_storage(config)
                except Exception as err:
                    logging.getLogger(logger_name).error("Config JSON has problems: {}".format(err))
                    if "codec" in str(err):
                        raise Exception("{} is not a utf-8 encoded file.".format(self.default_config_file_location))
                    raise err

        # PermissionError is an IOError child. Handle before IOError.
        except PermissionError as err:
            raise err
        except IOError:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self.default_config_file_location)

        return config

    def save_storage(self, updated_config):

        self.create_config_file_if_missing()

        with open(self.default_config_file_location, "w") as write_file:
            json.dump(updated_config, write_file, indent=4, sort_keys=True)

    def get(self, key: ConfigKeys):
        config = self.read_storage()

        return config.get(key.value)

    def set(self, key: ConfigKeys, value):
        config = self.read_storage()
        config[key.value] = value

        self.save_storage(config)

        return config

    def delete(self, key: ConfigKeys):
        config = self.read_storage()

        kv = key.value

        if kv in config:
            del config[kv]
            logging.getLogger(logger_name).debug("Removed key %s" % kv)
        else:
            logging.getLogger(logger_name).debug("No key %s was found in config" % kv)

        self.save_storage(config)

        return config

    def delete_all(self):
        config = self.read_storage()
        config.clear()

        self.save_storage(config)

        return config

    def contains(self, key: ConfigKeys):
        config = self.read_storage()

        return key.value in config

    def create_config_file_if_missing(self):
        if not os.path.exists(self.default_config_file_location):

            f = open(self.default_config_file_location, "w+")
            f.write(json.dumps({}))
            f.close()

            # Make sure the new config file has the correct mode.
            set_config_mode(self.default_config_file_location)

    def is_empty(self):
        config = self.read_storage()

        return not config


class InMemoryKeyValueStorage(KeyValueStorage):
    """ File based implementation of the key value storage"""

    def __init__(self, config=None):

        self.config = {}

        if config is None:
            config = {}

        elif isinstance(config, str):

            if InMemoryKeyValueStorage.is_base64(config):
                # Decode if config json was provided as base64 string
                config = utils.base64_to_string(config)

            config = json_to_dict(config)
            if not config:
                raise exceptions.KeeperError("Could not load config data. Json text size: %s" % str(len(config)))

        for key in config:
            self.config[ConfigKeys.get_enum(key)] = config[key]

    def read_storage(self):

        # To match what FileKeyValueStorage does, we need to return the enum values as keys instead
        # of the enum keys
        dict_config = {}
        for enum_key in self.config:
            dict_config[enum_key.value] = self.config[enum_key]
        return dict_config

    def save_storage(self, updated_config):
        pass

    def get(self, key: ConfigKeys):
        return self.config.get(key)

    def set(self, key: ConfigKeys, value):
        self.config[key] = value

    def delete(self, key: ConfigKeys):
        self.config.pop(key, None)

    def delete_all(self):
        self.config = {}

    def contains(self, key: ConfigKeys):
        return key in self.config

    @staticmethod
    def is_base64(s):
        try:
            return base64.b64encode(base64.b64decode(s)) == str.encode(s)
        except (Exception,):
            return False



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import base64
import datetime
import hashlib
import hmac
import json
import logging
import os
import sys
import random
import string
import time
from json import JSONDecodeError
from sys import platform as _platform
from typing import Optional, Tuple
from urllib import parse
import subprocess
import stat
from distutils.util import strtobool

from keeper_secrets_manager_core.keeper_globals import logger_name

ENCODING = 'UTF-8'
SPECIAL_CHARACTERS = '''"!@#$%()+;<>=?[]{}^.,'''
DEFAULT_PASSWORD_LENGTH = 32


def get_os():
    if _platform.lower().startswith("linux"):
        return "linux"
    elif _platform.lower().startswith("darwin"):
        return "macOS"
    # elif _platform.lower().startswith("win32"):
    #     return "win32"
    # elif _platform.lower().startswith("win64"):
    #     return "win64"
    else:
        return _platform


def bytes_to_string(b):
    return b.decode(ENCODING)


def bytes_to_int(b):
    return int.from_bytes(b, byteorder='big')


def bytes_to_base64(b):
    return base64.b64encode(b).decode()


def base64_to_bytes(s):
    return base64.urlsafe_b64decode(s)


def base64_to_string(b64s):
    return base64.b64decode(b64s).decode('UTF-8')


def string_to_bytes(s):
    return s.encode(ENCODING)


def url_safe_str_to_bytes(s):
    b = base64.urlsafe_b64decode(s + '==')
    return b


def url_safe_str_to_int(s):
    b = url_safe_str_to_bytes(s)
    return bytes_to_int(b)


def generate_random_bytes(length):
    return os.urandom(length)


def dict_to_json(dictionary):
    return json.dumps(dictionary, indent=4)


def json_to_dict(json_str):

    try:
        resp = json.loads(json_str)
    except JSONDecodeError as jsonDecErr:
        logging.getLogger(logger_name).warning(jsonDecErr)
        resp = None

    return resp


def now_milliseconds():
    return int(time.time() * 1000)


class TotpCode:

    def __init__(self, code, time_left, period):
        self.code = code
        self.time_left = time_left
        self.period = period


def get_totp_code(url):
    # type: (str) -> TotpCode or None
    comp = parse.urlparse(url)
    if comp.scheme != 'otpauth':
        raise ValueError('Not an otpauth URI')

    if comp.scheme == 'otpauth':
        secret = None
        algorithm = 'SHA1'
        digits = 6
        period = 30
        counter = 0

        # parse URL query string
        for k, v in parse.parse_qsl(comp.query):
            if k == 'secret':
                secret = v
            elif k == 'algorithm':
                algorithm = v
            elif k == 'digits':
                digits = int(v) if v.isnumeric() and int(v) > 0 else digits
            elif k == 'period':
                period = int(v) if v.isnumeric() and int(v) > 0 else period
            elif k == 'counter':
                counter = int(v) if v.isnumeric() and int(v) > 0 else counter

        # validate parameters
        if not secret:
            raise ValueError('TOTP secret not found in URI')

        hash = hashlib.sha1
        algorithm = algorithm.upper()
        if algorithm == 'SHA1':
            hash = hashlib.sha1
        elif algorithm == 'SHA256':
            hash = hashlib.sha256
        elif algorithm == 'SHA512':
            hash = hashlib.sha512
        else:
            raise ValueError('Invalid value "{0}" for TOTP algorithm, must be SHA1, SHA256 or SHA512'.format(algorithm))

        if digits not in [6, 7, 8]:
            raise ValueError('TOTP Digits may only be 6, 7, or 8')

        tm_base = counter if counter > 0 else int(datetime.datetime.now().timestamp())
        tm = tm_base / period
        reminder = len(secret) % 8
        if reminder in {2, 4, 5, 7}:
            padding = '=' * (8 - reminder)
            secret += padding
        key = base64.b32decode(secret, casefold=True)
        msg = int(tm).to_bytes(8, byteorder='big')
        hm = hmac.new(key, msg=msg, digestmod=hash)
        digest = hm.digest()

        offset = digest[-1] & 0x0f
        base = bytearray(digest[offset:offset + 4])
        base[0] = base[0] & 0x7f
        code_int = int.from_bytes(base, byteorder='big')
        code = str(code_int % (10 ** digits)).zfill(digits)
        elapsed = tm_base % period  # time elapsed in current period in seconds
        ttl = period - elapsed  # time to live in seconds

        return TotpCode(code, ttl, period)


#  password generation
def random_sample(sample_length=0, sample_string=''):
    use_secrets = False
    try:
        # Older version of Python (before 3.6) don't have this module.
        # If not installed, fall back to the original version of the code
        import secrets
        logging.debug("module 'secrets' is installed")
        use_secrets = True
    except ModuleNotFoundError:
        logging.warning("module 'secrets' is not installed")

    sample = ''
    for _ in range(sample_length):
        if use_secrets:
            sample += secrets.choice(sample_string)
        else:
            pos = int.from_bytes(os.urandom(2), 'big') % len(sample_string)
            sample += sample_string[pos]

    return sample


def get_windows_user_sid_and_name(logger=None):

    # Get the current user. Subprocess will run CMD. Getting from os.environ.get("USERNAME") can be flaky.
    user_output = subprocess.run(["whoami"], capture_output=True)
    if user_output.stdout is None:
        if logger is not None:
            logger.info("Cannot get current window user via 'whoami'")
        return None, None

    # Get only the username, not the machine
    user_parts = user_output.stdout.decode().strip().split("\\")
    user = user_parts[1] if len(user_parts) == 2 else user_parts[0]

    # Get the SID for the user. The 'where' command doesn't seem to work :(, so get them all.
    sid_output = subprocess.run(["wmic", "useraccount", "get",  "name,sid"], capture_output=True)
    if sid_output.stdout is None:
        if logger is not None:
            logger.info("Cannot user account list for SID lookup")
        return None, None

    sid = None
    if sid_output.stdout is not None:
        lines = sid_output.stdout.decode().split("\n")
        first_line = lines[0].lower()
        lines = lines[1:]
        index = first_line.index("sid")
        for line in lines:
            if line.lower().startswith(user) is True:
                sid = line[index:].strip()
                break
    if sid is None:
        if logger is not None:
            logger.info("Cannot find the SID for user " + user)

    return sid, user


def set_config_mode(file, logger=None):

    # Allow the user skip locking down the configuration file's mode.
    if bool(strtobool(os.environ.get("KSM_CONFIG_SKIP_MODE", "FALSE"))) is False:
        # For Windows, use icacls. cacls is obsolete.

        if _platform.lower().startswith("win") is True:

            sid, user = get_windows_user_sid_and_name()

            # https://stackoverflow.com/questions/5264595/windows-chmod-600
            # https://github.com/PowerShell/Win32-OpenSSH/issues/132

            # Remove mode inherited by the directory.
            # Remove everyone's access
            # Grant the current user full access
            # Allow Administrators full access
            commands = [
                'icacls.exe "{}" /reset'.format(file),
                'icacls.exe "{}" /inheritance:r'.format(file),
                'icacls.exe "{}" /remove:g Everyone:F'.format(file),
                'icacls.exe "{}" /grant:r Administrators:F'.format(file),
                'icacls.exe "{}" /grant:r *{}:F'.format(file, sid),
            ]
            for command in commands:
                if logger is not None:
                    logger.debug("Set Mode Command " + command)
                output = subprocess.run(command, capture_output=True)
                if output.stderr is not None and "Access is denied" in output.stderr.decode():
                    raise Exception("Access denied to configuration file {}.".format(file))
                if output.stdout is not None and "Failed processing 0 files" not in output.stdout.decode():
                    message = "Could not change the ACL for file '{}'. Set the environmental variable " \
                              "'KSM_CONFIG_SKIP_MODE' to 'TRUE' to skip setting the ACL mode.".format(file)
                    if output.stderr is not None:
                        message += ": " + output.stderr.decode()
                    else:
                        message += "."

                    raise Exception(message)
        else:
            # In Linux/MacOs get file permissions to 0600.
            os.chmod(file, stat.S_IREAD | stat.S_IWRITE)


def check_config_mode(file, color_mod=None, logger=None):

    # If we are skipping setting the mode, skip checking.
    if bool(strtobool(os.environ.get("KSM_CONFIG_SKIP_MODE", "FALSE"))) is False:

        # For Windows, use icacls. cacls is obsolete.
        if _platform.lower().startswith("win") is True:

            # If this doesn't error out, then we know the file exists
            output = subprocess.run(["icacls.exe", file], capture_output=True)
            if output.stderr is not None:
                if "Access is denied" in output.stderr.decode():
                    raise PermissionError("Access denied to configuration file {}.".format(file))
                if "cannot find" in output.stderr.decode():
                    raise FileNotFoundError("Cannot find configuration file {}.".format(file))

            # Try to access the file. If it now can't be found, it's a permission problem.
            try:
                with open(file, "r") as fh:
                    fh.close()
            except (FileNotFoundError, PermissionError):
                raise PermissionError("Access denied to configuration file {}.".format(file))

            if bool(strtobool(os.environ.get("KSM_CONFIG_SKIP_MODE_WARNING", "FALSE"))) is False:
                # We need to figure out who we are. subprocess run will use cmd

                sid, user = get_windows_user_sid_and_name()
                if sid is not None:
                    allowed_users = [user.lower(), "Administrators".lower()]
                    for line in output.stdout.decode().split("\n"):
                        parts = line[len(file):].split(":")
                        if len(parts) == 2:
                            found_user = parts[0].split("\\").pop()
                            if found_user.lower() not in allowed_users:

                                message = "The config file mode is too open for '{}'. Use `icacls` to remove access " \
                                          "for other users and groups.\n\n".format(file)
                                message += '> icacls.exe "{}" /reset\n'.format(file)
                                message += '> icacls.exe "{}" /inheritance:r\n'.format(file)
                                message += '> icacls.exe "{}" /remove:g Everyone:F\n'.format(file)
                                message += '> icacls.exe "{}" /grant:r Administrators:F\n'.format(file)
                                message += '> icacls.exe "{}" /grant:r *{}:F\n'.format(file, sid)
                                message += "\nTo disable this check, set the environmental variable " \
                                           "'KSM_CONFIG_SKIP_MODE_WARNING' to 'TRUE'."
                                if color_mod is not None:
                                    message = color_mod.Fore.RED + message + color_mod.Style.RESET_ALL

                                print(message, file=sys.stderr)
                                # Prevent multiple nagging per execution.
                                os.environ["KSM_CONFIG_SKIP_MODE_WARNING"] = "TRUE"
                                break
        else:
            # Can the user read the file? First check if the file exists. If it does, os.access might throw
            # and exception about it not existing. This mean we don't have access.
            if os.path.exists(file) is True:
                try:
                    if os.access(file, os.R_OK) is False:
                        raise PermissionError("Access denied to configuration file {}.".format(file))
                except FileNotFoundError:
                    raise PermissionError("Access denied to configuration file {}.".format(file))

            # Allow user to skip being nagged by warning message.
            if bool(strtobool(os.environ.get("KSM_CONFIG_SKIP_MODE_WARNING", "FALSE"))) is False:
                mode = oct(os.stat(file).st_mode)
                # Make sure group and user have no rights. Allow owner to have anything.
                if mode[-2:] != "00":
                    print("The config file mode, {}, is too open. "
                          "It is recommended to execute 'chmod 0600 {}' to remove group and user "
                          "access. To disable this warning, set the environment variable "
                          "'KSM_CONFIG_SKIP_MODE_WARNING' to 'TRUE'.".format(mode[-4:], file), file=sys.stderr)
def generate_password(length: int = DEFAULT_PASSWORD_LENGTH,
                      lowercase: Optional[int] = None,
                      uppercase: Optional[int] = None,
                      digits: Optional[int] = None,
                      special_characters: Optional[int] = None,
                      special_characterset: str = SPECIAL_CHARACTERS):
    """
    generate_password generates a new password of specified minimum length
    with specified number of uppercase, lowercase, digits and special characters.
    Note: If all character groups are unspecified or all have exact zero length
    then password characters are chosen from all groups uniformly at random.
    Note: If all charset lengths are negative or 0 but can't reach the minimum length
    then all exact/negative charset lengths will be treated as minimum number of characters instead.

    :param length: minimum password length - default: 32
    :param lowercase: minimum number of lowercase characters if positive, exact if 0 or negative
    :param uppercase: minimum number of uppercase characters if positive, exact if 0 or negative
    :param digits: minimum number of digits if positive, exact if 0 or negative
    :param special_characters: minimum number of special characters if positive, exact if 0 or negative
    :param special_characterset: string containing custom set of special characters to pick from
    :return: generated password string
    """

    counts = (lowercase, uppercase, digits, special_characters)
    sum_categories = sum((abs(i) if isinstance(i, int) else 0) for i in counts)

    # If all lengths are exact/negative but don't reach minimum length - convert to minimum/positive lengths
    num_exact_counts = sum(1 for i in counts if isinstance(i, int) and i <= 0)
    if len(counts) == num_exact_counts and sum_categories < length:
        lowercase = abs(lowercase) if isinstance(lowercase, int) and lowercase < 0 else lowercase
        uppercase = abs(uppercase) if isinstance(uppercase, int) and uppercase < 0 else uppercase
        digits = abs(digits) if isinstance(digits, int) and digits < 0 else digits
        special_characters = abs(special_characters) if isinstance(special_characters, int) and special_characters < 0 else special_characters
        logging.getLogger(logger_name).warning("Bad charset lengths - converting exact lengths to minimum lengths")

    extra_count = length - sum_categories if length > sum_categories else 0
    extra_chars = ''
    if lowercase is None or isinstance(lowercase, int) and lowercase > 0:
        extra_chars += string.ascii_lowercase
    if uppercase is None or isinstance(uppercase, int) and uppercase > 0:
        extra_chars += string.ascii_uppercase
    if digits is None or isinstance(digits, int) and digits > 0:
        extra_chars += string.digits
    if special_characters is None or isinstance(special_characters, int) and special_characters > 0:
        extra_chars += special_characterset
    if extra_count > 0 and not extra_chars:
        extra_chars = string.ascii_lowercase + string.ascii_uppercase + string.digits + special_characterset

    category_map = [
        (abs(lowercase) if isinstance(lowercase, int) else 0, string.ascii_lowercase),
        (abs(uppercase) if isinstance(uppercase, int) else 0, string.ascii_uppercase),
        (abs(digits) if isinstance(digits, int) else 0, string.digits),
        (abs(special_characters) if isinstance(special_characters, int) else 0, special_characterset),
        (extra_count, extra_chars)
    ]

    password_list = []
    for count, chars in category_map:
        password_list.extend(list(random_sample(count, chars)))
    random.shuffle(password_list)
    return ''.join(password_list)



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
import json
import mimetypes
import os
from datetime import datetime
from pathlib import Path

import requests
from keeper_secrets_manager_core.crypto import CryptoUtils
from keeper_secrets_manager_core.exceptions import KeeperError
from keeper_secrets_manager_core import utils, helpers


class Record:

    def __init__(self, record_dict, secret_key):
        self.uid = ''
        self.title = ''
        self.type = ''
        self.files = []
        self.raw_json = None
        self.dict = {}
        self.password = None
        self.revision = None
        self.is_editable = None

        self.uid = record_dict.get('recordUid')

        if 'recordKey' in record_dict and record_dict.get('recordKey'):
            # Folder Share
            record_key_encrypted_str = record_dict.get('recordKey')
            record_key_encrypted_bytes = utils.base64_to_bytes(record_key_encrypted_str) if \
                record_key_encrypted_str else None

            self.record_key_bytes = CryptoUtils.decrypt_aes(record_key_encrypted_bytes, secret_key)
        else:
            # Single Record Share
            self.record_key_bytes = secret_key

        record_encrypted_data = record_dict.get('data')
        record_data_json = CryptoUtils.decrypt_record(record_encrypted_data, self.record_key_bytes)

        self.raw_json = record_data_json
        self.dict = utils.json_to_dict(self.raw_json)
        self.title = self.dict.get('title')
        self.type = self.dict.get('type')
        self.revision = record_dict.get('revision')
        self.is_editable = record_dict.get("isEditable")

        # files
        if record_dict.get('files'):
            for f in record_dict.get('files'):

                file = KeeperFile(f, self.record_key_bytes)

                self.files.append(file)

        # password (if `login` type)
        if self.type == 'login':

            fields = self.dict.get('fields')

            password_field = next((item for item in fields if item["type"] == "password"), None)

            # If the password field exists and there is a value in the array, then set the password.
            if password_field is not None and len(password_field.get('value', [])) > 0:
                self.password = password_field.get('value')[0]

    def find_file_by_title(self, title):
        """Finds file by file title"""

        found_file = next((f for f in self.files if f.title == title), None)

        return found_file

    def download_file_by_title(self, title, path):

        found_file = self.find_file_by_title(title)

        if not found_file:
            raise KeeperError("File %s not found" % title)

        found_file.save_file(path)

    def __str__(self):
        return '[Record: uid=%s, type: %s, title: %s, files count: %s]' % (self.uid, self.type, self.title,
                                                                           str(len(self.files)))

    def _update(self):

        """ Take the values in the diction and update the attributes and raw JSON
        """

        self.dict["title"] = self.title
        self.dict["type"] = self.type

        # Find the password in the field and update the password attribute
        password_field = next((item for item in self.dict["fields"] if item["type"] == "password"), None)
        if password_field is not None and len(password_field.get('value', [])) > 0:
            self.password = password_field.get('value')[0]

        self.raw_json = utils.dict_to_json(self.dict)

    @staticmethod
    def _value(values, single):

        if single is True:
            if values is None or len(values) == 0:
                return None
            return values[0]
        return values

    @staticmethod
    def _field_search(fields, field_key):

        """ This is a generic field search that returns the field

        It will work for for both standard and custom fields. It
        returns the field as a dictionary.
        """

        # First check in the field_key matches any labels. Label matching is case sensitive.
        found_item = None
        for item in fields:
            if item.get("label") is not None and item.get("label") == field_key:
                found_item = item
                break
        # If the label was not found, check the field type. Field type is case insensitive.
        if found_item is None:
            for item in fields:
                if item.get("type").lower() == field_key.lower():
                    found_item = item
                    break

        return found_item

    def get_standard_field(self, field_type):
        return self._field_search(fields=self.dict.get('fields', []), field_key=field_type)

    def get_standard_field_value(self, field_type, single=False):
        field = self.get_standard_field(field_type)
        if field is None:
            raise ValueError("Cannot find standard field {} in record".format(field_type))
        return Record._value(field.get("value", []), single)

    def set_standard_field_value(self, field_type, value):
        field = self.get_standard_field(field_type)
        if field is None:
            raise ValueError("Cannot find standard field {} in record".format(field_type))
        if type(value) is not list:
            value = [value]
        field["value"] = value
        self._update()

    def get_custom_field(self, field_type):
        return self._field_search(fields=self.dict.get('custom', []), field_key=field_type)

    def get_custom_field_value(self, field_type, single=False):
        field = self.get_custom_field(field_type)
        if field is None:
            raise ValueError("Cannot find custom field {} in record".format(field_type))
        return Record._value(field.get("value", []), single)

    def set_custom_field_value(self, field_type, value):
        field = self.get_custom_field(field_type)
        if field is None:
            raise ValueError("Cannot find custom field {} in record".format(field_type))
        if type(value) is not list:
            value = [value]
        field["value"] = value
        self._update()

    # TODO: Deprecate this for better getter and setters
    def field(self, field_type, value=None, single=False):

        """ Getter and setter for standard fields

        A getter operation is performed when the 'value' parameter is not passed. For example, this would
        return the value.

            record.field("login")

        A setter operation is performed when a 'value'  parameter is passed. For example, this would set
        the value in the field.

            record.field("login", value="My New Value")
        """

        field = self._field_search(fields=self.dict.get('fields', []), field_key=field_type)

        if field is None:
            raise ValueError("Cannot find the field for {}".format(field_type))

        if value is None:
            value = Record._value(field["value"], single)
        else:
            if type(value) is not list:
                value = [value]
            field["value"] = value
            self._update()

        return value

    # TODO: Deprecate this for better getter and setters
    def custom_field(self, label=None, value=None, field_type=None, single=False):

        custom_field = None
        if label is not None:
            custom_field = self._field_search(fields=self.dict.get('custom', []), field_key=label)
            if custom_field is None and field_type is not None:
                custom_field = self._field_search(fields=self.dict.get('custom', []), field_key=field_type)

        if custom_field is None:
            raise ValueError("Cannot find the custom field label='{}', field type='{}'.".format(label, field_type))

        if value is None:
            value = Record._value(custom_field["value"], single)
        else:
            if type(value) is not list:
                value = [value]
            custom_field["value"] = value
            self._update()

        return value

    def print(self):

        print("===")
        print("Title: {}".format(self.title))
        print("UID:   {}".format(self.uid))
        print("Type:  {}".format(self.type))
        print("")
        print("Fields")
        print("------")

        for item in self.dict.get('fields'):
            if item["type"] in ["fileRef", "oneTimeCode"]:
                continue
            print("{} : {}".format(item["type"], ", ".join(item["value"])))

        print("")
        print("Custom Fields")
        print("------")
        for item in self.dict.get('custom', []):
            print("{} ({}) : {}".format(item["label"], item["type"], ", ".join(item["value"])))


class Folder:

    def __init__(self, folder, secret_key):

        self.uid = ''
        self.records = []

        if not folder.get('folderUid'):
            raise Exception("Not a folder")

        folder_uid = folder.get('folderUid')
        folder_key_enc = folder.get('folderKey')
        folder_key = CryptoUtils.decrypt_aes(utils.base64_to_bytes(folder_key_enc), secret_key)
        folder_records = folder.get('records')

        self.uid = folder_uid
        self.key = folder_key

        for r in folder_records:

            record = Record(r, folder_key)
            self.records.append(record)


class KeeperFile:

    def __init__(self, f, record_key_bytes):

        self.file_key = ''
        self.meta_dict = None

        self.file_data = None

        self.name = ''
        self.title = ''
        self.type = ''
        self.last_modified = 0
        self.size = 0

        self.f = f
        self.record_key_bytes = record_key_bytes

        # Set file metadata

        meta = self.__get_meta()

        self.title = meta.get('title')
        self.name = meta.get('name')
        self.type = meta.get('type')
        self.last_modified = meta.get('lastModified')
        self.size = meta.get('size')

    def __decrypt_file_key(self):
        file_key_encrypted_base64 = self.f.get('fileKey')
        file_key_encrypted = utils.base64_to_bytes(file_key_encrypted_base64)
        file_key = CryptoUtils.decrypt_aes(file_key_encrypted, self.record_key_bytes)
        return file_key

    def __get_meta(self):
        """
        Returns file metadata dictionary (file name, title, size, type, etc.)
        """
        if not self.meta_dict:
            file_key = self.__decrypt_file_key()

            meta_json = CryptoUtils.decrypt_aes(utils.base64_to_bytes(self.f.get('data')), file_key)

            self.meta_dict = utils.json_to_dict(meta_json)

        return self.meta_dict

    def get_file_data(self):
        """
        Return decrypted raw file data
        """
        if not self.file_data:    # cached if nothing
            file_key = self.__decrypt_file_key()
            file_url = self.f.get('url')

            rs = requests.get(file_url)

            file_encrypted_data = rs.content

            self.file_data = CryptoUtils.decrypt_aes(file_encrypted_data, file_key)

        return self.file_data

    def save_file(self, path, create_folders=False):
        """
        Save decrypted file data to the provided path
        """

        if create_folders:
            os.makedirs(os.path.dirname(path), exist_ok=True)

        file_data = self.get_file_data()

        dir_path = os.path.dirname(os.path.abspath(path))

        if not os.path.exists(dir_path):
            raise KeeperError("No such file or directory %s\nConsider adding `create_folders=True` to `save_file()` "
                              "method " % path)

        file = open(path, "wb")

        file.write(file_data)
        file.close()

        return True

    def __str__(self):
        return "[KeeperFile - name: %s, title: %s]" % (self.name, self.title)


class KeeperFileUpload:

    def __init__(self, name=None, title=None, mime_type=None, data=None):
        self.Name = name
        self.Title = title
        self.Type = mime_type
        self.Data = data

    @staticmethod
    def from_file(path, file_name=None, file_title=None, mime_type=None):

        """Helper method to get Keeper File Upload object from the file path"""
        file_name = file_name if file_name else Path(path).name
        file_title = file_title if file_title else file_name

        if not mime_type:
            mime_type = mimetypes.guess_type(path)[0]

            if not mime_type:
                # fall back to `application/octet-stream` if type was not determined
                mime_type = 'application/octet-stream'

        in_file = open(path, 'rb')  # opening for [r]eading as [b]inary
        file_bytes_data = in_file.read()

        file_upload = KeeperFileUpload(name=file_name, title=file_title, mime_type=mime_type, data=file_bytes_data)

        return file_upload


class KeeperFileData:

    def __init__(self):
        self.name = None
        self.size = None
        self.title = None
        self.lastModified = None
        self.type = None


VALID_RECORD_FIELDS = ['login', 'password', 'url', 'fileRef', 'oneTimeCode', 'otp', 'name', 'birthDate', 'date',
                       'expirationDate', 'text', 'securityQuestion', 'multiline', 'email', 'cardRef', 'addressRef',
                       'pinCode', 'phone', 'secret', 'note', 'accountNumber', 'paymentCard', 'bankAccount', 'keyPair',
                       'host', 'address', 'licenseNumber']


class RecordField:

    def __init__(self, field_type=None, value=None, label=None, required=None, enforceGeneration=None,
                 privacyScreen=None, complexity=None):

        self.type = field_type

        if isinstance(value, list):
            self.value = value
        else:
            self.value = [value] if value else []

        if label:
            self.label = label
        if required:
            self.required = required
        if enforceGeneration:
            self.enforceGeneration = enforceGeneration
        if privacyScreen:
            self.privacyScreen = privacyScreen
        if complexity:
            self.complexity = complexity


class RecordCreate:

    def __init__(self, record_type, title):

        self.record_type = record_type
        self.title = title
        self.notes = None
        self.fields = None
        self.custom = None

    def _validate(self):

        # Validate title
        if not isinstance(self.title, str):
            raise KeeperError(f"Record title should be a string. Provided type {type(self.title)}")

        # Validate notes
        if self.notes and not isinstance(self.notes, str):
            raise KeeperError(f"Record notes should be a string. Provided type {type(self.notes)}")

        field_type_errors = []
        field_value_errors = []

        if self.fields:
            for field in self.fields:
                field_type = field.type
                if field_type not in VALID_RECORD_FIELDS:
                    field_type_errors.append(field_type)

                field_value = field.value
                if not isinstance(field_value, list):
                    field_value_errors.append(field_type)

            # Validate field types - only legit field type names
            if len(field_type_errors) > 0:
                raise KeeperError(f"Following field types are not allowed [{', '.join(field_type_errors)}]. "
                                  f"Allowed field types are [{', '.join(VALID_RECORD_FIELDS)}]")

            # Validate field values - arrays only
            if len(field_value_errors) > 0:
                raise KeeperError(f"Fields with the following types are of a list type [{', '.join(field_value_errors)}]. "
                                  f"Make sure that those fields have values as a list type")

    def to_dict(self):

        self._validate()

        rec_dict = {
            'type': self.record_type,
            'title': self.title,
            'fields': self.fields,
        }

        if self.notes:
            rec_dict['notes'] = self.notes

        if self.custom:
            rec_dict['custom'] = self.custom

        return helpers.obj_to_dict(rec_dict)

    def to_json(self):

        json_object = json.dumps(self.to_dict(), indent=4)
        return json_object


class AppData:
    """
    Application info
    """
    def __init__(self, title="", app_type=""):
        self.title = title
        self.app_type = app_type


class SecretsManagerResponse:

    """
    Server response contained details about the application and the records
    that were requested to be returned
    """
    def __init__(self):
        self.appData = None
        # self.encryptedAppKey = None
        # self.appOwnerPublicKey = None
        self.folders = None
        self.records = None
        self.expiresOn = None
        self.warnings = None
        self.justBound = False

    def expires_on_str(self, date_format='%Y-%m-%d %H:%M:%S'):
        """
        Retrieve string formatted expiration date
        """
        return datetime.fromtimestamp(self.expiresOn/1000).strftime(date_format)


class SecretsManagerAddFileResponse:

    def __init__(self):
        self.url = None
        self.parameters = None
        self.successStatusCode = None



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import base64
import uuid
import json
import time
from collections import deque
from requests import Response as RequestResponse
from .keeper_globals import keeper_public_keys
import string
import random

from keeper_secrets_manager_core.crypto import CryptoUtils
from keeper_secrets_manager_core.configkeys import ConfigKeys
from keeper_secrets_manager_core.dto.payload import KSMHttpResponse


class ResponseQueue:

    """Queue up response

    The is a FIFO queue. The queue can be loaded with mock Response instance that
    will be shift off when a request to get_response called.

    """

    def __init__(self, client, post_method=None):

        self.client = client

        # If the post_method is not defined, default to a built in one. All it will do is
        # return the next response in thr queue.
        if post_method is None:
            post_method = self.auto_responder_patch
        self.post_method = post_method

        self.queue = deque()

    def add_response(self, response):
        response.client = self.client
        if self.post_method is not None:
            response.patch_post_query(self.post_method)
        self.queue.append(response)

    def get_response(self, transmission_key):
        if len(self.queue) == 0:
            raise ValueError("Not enough queued responses. Cannot get response.")
        response = self.queue.popleft()
        return response.instance(transmission_key)

    def auto_responder_patch(self, _, transmission_key, _two, _three):
        return self.get_response(transmission_key)


class Response:

    """Mock a response from Secrets Manager Service

    secrets_manager_instance = SecretManger()
    mock_res = MockResponse(client=secrets_manager_instance)
    record = mock_res.add_record(title="My Record")
    record.login = "My Login"
    record.password = "My Password"
    record.add_file(name="My File 1")
    record.add_file(name="My File 2")

    response = mock_res.response()

    """

    def __init__(self, client=None, post_method=None, content=None, status_code=200, reason="OK", flags=None):

        """
        Flags is a dictionary where we can turn on/off stuff in the response to fake qwarks from other application and
        how they might store the data.

        * prune_custom_fields - If there is no custom fields, the 'custom' key is removed from the JSON.

        """

        self.client = client

        self.records = {}
        self.folders = {}

        self.content = content
        self.status_code = status_code
        self.reason = reason
        self.flags = flags

        # Monkey patch the _post_query method with a method that will return a
        # requests Response obj.
        if post_method is not None:
            self.patch_post_query(post_method)

        self.headers = {
            "Server": "keeper",
            "Content-Type": "application/octet-stream",
            "Connection": "keep-alive",
            "X-Frame-Options": "DENY",
            "Strict-Transport-Security": "max-age=31536000; includeSubDomains; preload",
            "X-Content-Type-Options": "nosniff",
            "X-XSS-Protection": "1; mode=block",
            "Expect-CT": "max-age=10, report-uri=\"https://keepersecurity.report-uri.com/r/d/ct/reportOnly\"",
            "Accept-Ranges": "bytes"
        }

    def patch_post_query(self, patch):
        """Patch keeper_secrets_manager.core.SecretsManager._post_query

        The patch is a method that accept three arguments.

          * path - Endpoint Path.
          * context - Context instance.
          * payload_and_signature - Dictionary containing encrypted data and signature.

        When SecretsManager posts a message to the Secrets Manager services, it will call the patch method
        instead. The patch method needs to return an instance of the requests Response.

        """

        if self.client is None:
            raise ValueError("The secrets manager client has not been set.")

        # def _post_function(url, transmission_key, encrypted_payload_and_signature, verify_ssl_certs=True):
        # We need the transmission_key.key to encrypt the content
        self.client.post_function = patch

    def dump(self, secret, flags=None):

        return {
            "encryptedAppKey": None,
            # "appOwnerPublicKey" ????????,
            "folders": [self.folders[uid].dump(secret=secret, flags=flags) for uid in self.folders],
            "records": [self.records[uid].dump(secret=secret, flags=flags) for uid in self.records]
        }

    def instance(self, transmission_key):

        """ Return a requests Response instance filled in with mock response message

        res = Response.instance(context)

        The method requires an instance of keeper_secrets_manager_core.dto.payload.Context since that
        information on how to encrypt the response message.

        """

        res = RequestResponse()
        for key in self.headers:
            res.headers[key] = self.headers[key]
        res.headers["Date"] = time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())

        # If canned content has not be set, the create content from records/folders.
        if self.content is None:
            app_key = self.client.config.get(ConfigKeys.KEY_APP_KEY)
            app_key = base64.urlsafe_b64decode(app_key + "==")
            json_str = json.dumps(self.dump(secret=app_key, flags=self.flags))
            content = CryptoUtils.encrypt_aes(json_str.encode(), transmission_key.key)
            res._content = content
            res.headers["Content-Length"] = str(len(content))

            res.status_code = 200
            res.reason = "OK"
        # Else return the canned content. This is useful to mock errors that return plain or json text.
        else:
            content = self.content
            if type(content) is str:
                content = content.encode()
            res._content = content
            res.status_code = self.status_code
            res.reason = self.reason

        return KSMHttpResponse(res.status_code, res.content, res)

    def add_record(self, title=None, record_type=None, uid=None, record=None, keeper_record=None):

        if keeper_record is not None:
            record = Record.convert_keeper_record(keeper_record)
        elif record is None:
            record = Record(title=title, record_type=record_type, uid=uid)

        if isinstance(object, Record.__class__) is False:
            raise ValueError("Record being added to the response is not a "
                             "keeper_secrets_manager_core.mock.Record instance.")

        self.records[record.uid] = record
        return record

    def add_folder(self, uid=None, folder=None):

        if folder is None:
            folder = Folder(uid=uid)

        if isinstance(object, Folder.__class__) is False:
            raise ValueError("Folder being added to the response is not a "
                             "keeper_secrets_manager_core.mock.Folder instance.")

        self.folders[folder.uid] = folder
        return folder


class Folder:

    def __init__(self, uid=None):
        if uid is None:
            uid = str(uuid.uuid4())
        self.uid = uid
        self.records = {}

    def add_record(self, title=None, record_type=None, uid=None, record=None):

        if record is None:
            record = Record(record_type=record_type, uid=uid, title=title)

        if isinstance(object, Record.__class__) is False:
            raise ValueError("Record being added to the response is not a "
                             "keeper_secrets_manager_core.mock.Record instance.")

        self.records[record.uid] = record
        return record

    def dump(self, secret, flags=None):

        return {
            "folderUid": self.uid,
            "folderKey": base64.b64encode(CryptoUtils.encrypt_aes(secret, secret)).decode(),
            "records": [self.records[uid].dump(secret=secret, flags=flags) for uid in self.records]
        }


class File:

    def __init__(self, name, title=None, content_type=None, url=None, content=None, last_modified=None):
        self.uid = str(uuid.uuid4())
        self.secret_used = None

        self.name = name
        if title is None:
            title = self.name
        self.title = title
        if content_type is None:
            content_type = "text/plain"
        self.content_type = content_type
        if url is None:
            url = "http://localhost/{}".format(self.uid)
        self.url = url
        if content is None:
            content = "ABC123"
        self.content = content
        self.size = len(self.content)
        if last_modified is None:
            last_modified = int(time.time())
        self.last_modified = last_modified

    def downloadable_content(self):

        # The dump method will generate the content that the secrets manager would return. The
        # problem is we won't know the secret here. So the dump method needs to be run before
        # this method is called. The dump method will set the last/only secret used. We need to
        # encode the content with that secret.
        if self.secret_used is None:
            raise ValueError("The file has not be dump'd yet, Secret is unknown.")

        data = self.content
        if type(data) is not bytes:
            data = data.encode()

        return CryptoUtils.encrypt_aes(data, self.secret_used)

    def dump(self, secret, flags=None):

        # No special flags for download. Do this to make PEP8 happy for unused vars.
        if flags is not None:
            pass

        self.secret_used = secret
        d = {
            "name": self.name,
            "title": self.title,
            "size": self.size,
            "lastModified": self.last_modified,
            "type": self.content_type
        }
        data = json.dumps(d)
        file_data = {
            "fileUid": self.uid,
            "fileKey": base64.b64encode(CryptoUtils.encrypt_aes(secret, secret)).decode(),
            "data": base64.b64encode(CryptoUtils.encrypt_aes(data.encode(), secret)).decode(),
            "url": self.url,
            "thumbnailUrl": None
        }
        return file_data


class Record:

    no_label = "__NONE__"

    def __init__(self, record_type=None, uid=None, title=None):

        if uid is None:
            uid = str(uuid.uuid4())
        if record_type is None:
            record_type = "login"

        self.uid = uid
        self.record_type = record_type
        self.title = title
        self.is_editable = False
        self.files = {}

        self._fields = []
        self._custom_fields = []

    @staticmethod
    def convert_keeper_record(keeper_record):

        new_record = Record(
            record_type=keeper_record.type,
            uid=keeper_record.uid,
            title=keeper_record.title
        )
        for item in keeper_record.dict.get("fields"):
            new_record.field(
                label=item.get("label"),
                field_type=item["type"],
                value=item["value"]
            )
        for item in keeper_record.dict.get("custom", []):
            new_record.custom_field(
                label=item["label"],
                field_type=item["type"],
                value=item["value"]
            )
        # TODO - Add files
        return new_record

    @staticmethod
    def _field(field_type, value, label=None, required=None, privacy_screen=None):
        if isinstance(value, list) is False:
            value = [value]

        field = {
            "type": field_type,
            "value": value,
        }
        if label is not None:
            field["label"] = label
        if required is not None:
            field["required"] = required
        if privacy_screen is not None:
            field["privacyScreen"] = privacy_screen

        return field

    def field(self, field_type, value, label=None, required=None, privacy_screen=None):
        self._fields.append(
            self._field(field_type, value, label, required, privacy_screen)
        )

    def custom_field(self, label, value, field_type="text", required=None, privacy_screen=None):
        self._custom_fields.append(
            self._field(field_type, value, label, required, privacy_screen)
        )

    def add_file(self, name, title=None, content_type=None, url=None, content=None, last_modified=None):

        file = File(
            name=name,
            title=title,
            content_type=content_type,
            url=url,
            content=content,
            last_modified=last_modified
        )
        self.files[file.uid] = file
        return file

    def dump(self, secret, flags=None):

        fields = list(self._fields)

        # If no files, the JSON has null
        files = None
        if len(self.files) > 0:
            files = [self.files[uid].dump(secret=secret, flags=flags) for uid in self.files]
            fields.append({"type": "fileRef", "value": [uid for uid in self.files]})

        record_data = {
            "title": self.title,
            "type": self.record_type,
            "fields": fields,
            "custom": self._custom_fields
        }

        if flags is not None:
            # SecretsManager will not add a custom key if there is no custom fields. However the UI does.
            if flags.get("prune_custom_fields", False) is True and len(record_data["custom"]) == 0:
                record_data.pop("custom", None)
            # Remove fields if they do not have values.
            if flags.get("prune_empty_fields", False) is True:
                new_fields = []
                for field in record_data["fields"]:
                    if len(field.get("value")) > 0:
                        new_fields.append(field)
                record_data["fields"] = new_fields

        data = {
            "recordUid": self.uid,
            "recordKey": base64.b64encode(CryptoUtils.encrypt_aes(secret, secret)).decode(),
            "data": base64.b64encode(CryptoUtils.encrypt_aes(json.dumps(record_data).encode(), secret)).decode(),
            "isEditable": self.is_editable,
            "files": files
        }

        return data


class MockConfig:

    @staticmethod
    def make_config(skip_list=None, token=None, app_key=None, owner_key=None):

        if skip_list is None:
            skip_list = []

        if token is None:
            random_token = ''.join((random.choice(string.ascii_lowercase) for _ in range(28)))
            token = base64.urlsafe_b64encode(random_token.encode()).decode()
            token = token.replace("=", "")

        import keeper_secrets_manager_core.core as sm_core
        import keeper_secrets_manager_core.storage as sm_storage

        # Generate a fake hostname
        hostname = ''.join((random.choice(string.ascii_lowercase) for _ in range(10))) + ".com"

        sm_config = sm_storage.InMemoryKeyValueStorage()
        sm_core.SecretsManager(token=token, hostname=hostname, config=sm_config)

        if app_key is None:
            random_app_key = ''.join((random.choice(string.ascii_lowercase) for _ in range(32)))
            app_key = base64.b64encode(random_app_key.encode()).decode()
        sm_config.set(ConfigKeys.KEY_APP_KEY, app_key)

        if owner_key is None:
            owner_key = keeper_public_keys["7"]
        sm_config.set(ConfigKeys.KEY_OWNER_PUBLIC_KEY, owner_key)

        config = {}
        for key in ConfigKeys:
            if key.value in skip_list:
                continue
            if sm_config.contains(key) is True:
                config[key.value] = sm_config.get(key)

        return config

    @staticmethod
    def make_json(skip_list=None, token=None, app_key=None, config=None):
        if config is None:
            config = MockConfig.make_config(skip_list=skip_list, token=token, app_key=app_key)
        return json.dumps(config)

    @staticmethod
    def make_base64(skip_list=None, token=None, app_key=None, config=None, json_config=None):
        if json_config is None:
            json_config = MockConfig.make_json(skip_list=skip_list, token=token, app_key=app_key, config=config)
        return base64.b64encode(json_config.encode()).decode()



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import importlib_metadata
import re

logger_name = 'ksm'


def get_client_version(hardcode=False):
    """Get the version of the client

    The version # comes from metadata. In a unit test, there is metadata, so the version will
    be the defined major and 0.0 (ie 16.0.0.).

    If installed, the method can get the real version number. For the client version number
    we use the defined major and minor and revision numbers of the module version.

    For example, module version of 0.1.23 would create a client version would be 16.1.23.

    We also need to remove any alpha characters from the revision since Python allows that.

    """
    # Get the version of the keeper secrets manager core
    version_major = "16"
    version = "{}.4.0".format(version_major)

    # Allow the default version to be hard coded. If not build the client version from the module
    # version.
    if hardcode is False:
        try:
            ksm_version = importlib_metadata.version("keeper-secrets-manager-core")
            version_parts = ksm_version.split(".")
            version_minor = version_parts[1]
            version_revision = re.search(r'^\d+', version_parts[2]).group()
            version = "{}.{}.{}".format(version_major, version_minor, version_revision)
        except importlib_metadata.PackageNotFoundError:
            # In a unit test or development run, not an installed version. Just use the default.
            pass
        except Exception as err:
            raise Exception(err)
    return version


# Right now the client version is being hardcoded.
keeper_secrets_manager_sdk_client_id = "mp{}".format(get_client_version(hardcode=False))


keeper_public_keys = {
    '1': 'BK9w6TZFxE6nFNbMfIpULCup2a8xc6w2tUTABjxny7yFmxW0dAEojwC6j6zb5nTlmb1dAx8nwo3qF7RPYGmloRM',
    '2': 'BK9w6TZFxE6nFNbMfIpULCup2a8xc6w2tUTABjxny7yFmxW0dAEojwC6j6zb5nTlmb1dAx8nwo3qF7RPYGmloRM',
    '3': 'BK9w6TZFxE6nFNbMfIpULCup2a8xc6w2tUTABjxny7yFmxW0dAEojwC6j6zb5nTlmb1dAx8nwo3qF7RPYGmloRM',
    '4': 'BK9w6TZFxE6nFNbMfIpULCup2a8xc6w2tUTABjxny7yFmxW0dAEojwC6j6zb5nTlmb1dAx8nwo3qF7RPYGmloRM',
    '5': 'BK9w6TZFxE6nFNbMfIpULCup2a8xc6w2tUTABjxny7yFmxW0dAEojwC6j6zb5nTlmb1dAx8nwo3qF7RPYGmloRM',
    '6': 'BK9w6TZFxE6nFNbMfIpULCup2a8xc6w2tUTABjxny7yFmxW0dAEojwC6j6zb5nTlmb1dAx8nwo3qF7RPYGmloRM',

    '7': 'BK9w6TZFxE6nFNbMfIpULCup2a8xc6w2tUTABjxny7yFmxW0dAEojwC6j6zb5nTlmb1dAx8nwo3qF7RPYGmloRM',
    '8': 'BKnhy0obglZJK-igwthNLdknoSXRrGB-mvFRzyb_L-DKKefWjYdFD2888qN1ROczz4n3keYSfKz9Koj90Z6w_tQ',
    '9': 'BAsPQdCpLIGXdWNLdAwx-3J5lNqUtKbaOMV56hUj8VzxE2USLHuHHuKDeno0ymJt-acxWV1xPlBfNUShhRTR77g',
    '10': 'BNYIh_Sv03nRZUUJveE8d2mxKLIDXv654UbshaItHrCJhd6cT7pdZ_XwbdyxAOCWMkBb9AZ4t1XRCsM8-wkEBRg',
    '11': 'BA6uNfeYSvqagwu4TOY6wFK4JyU5C200vJna0lH4PJ-SzGVXej8l9dElyQ58_ljfPs5Rq6zVVXpdDe8A7Y3WRhk',
    '12': 'BMjTIlXfohI8TDymsHxo0DqYysCy7yZGJ80WhgOBR4QUd6LBDA6-_318a-jCGW96zxXKMm8clDTKpE8w75KG-FY',
    '13': 'BJBDU1P1H21IwIdT2brKkPqbQR0Zl0TIHf7Bz_OO9jaNgIwydMkxt4GpBmkYoprZ_DHUGOrno2faB7pmTR7HhuI',
    '14': 'BJFF8j-dH7pDEw_U347w2CBM6xYM8Dk5fPPAktjib-opOqzvvbsER-WDHM4ONCSBf9O_obAHzCyygxmtpktDuiE',
    '15': 'BDKyWBvLbyZ-jMueORl3JwJnnEpCiZdN7yUvT0vOyjwpPBCDf6zfL4RWzvSkhAAFnwOni_1tQSl8dfXHbXqXsQ8',
    '16': 'BDXyZZnrl0tc2jdC5I61JjwkjK2kr7uet9tZjt8StTiJTAQQmnVOYBgbtP08PWDbecxnHghx3kJ8QXq1XE68y8c',
    '17': 'BFX68cb97m9_sweGdOVavFM3j5ot6gveg6xT4BtGahfGhKib-zdZyO9pwvv1cBda9ahkSzo1BQ4NVXp9qRyqVGU'
}

keeper_servers = {
    'US': 'keepersecurity.com',
    'EU': 'keepersecurity.eu',
    'AU': 'keepersecurity.com.au',
    'GOV': 'govcloud.keepersecurity.us',
    'JP': 'keepersecurity.jp',
    'CA': 'keepersecurity.ca'
}



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

class KeeperError(Exception):

    def __init__(self, message):

        self.message = message

        super().__init__(self.message)


class KeeperAccessDenied(Exception):

    def __init__(self, message):

        self.message = message

        super().__init__(self.message)



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

class Context:

    def __init__(self, transmission_key, client_id, client_key):
        self.transmissionKey = transmission_key
        self.clientId = client_id
        self.clientKey = client_key


class TransmissionKey:

    def __init__(self, publicKeyId, key, encryptedKey):
        self.publicKeyId = publicKeyId
        self.key = key
        self.encryptedKey = encryptedKey


class GetPayload:

    def __init__(self):
        self.clientVersion = None
        self.clientId = None
        self.publicKey = None
        self.requestedRecords = None


class CreatePayload:

    def __init__(self):
        self.clientVersion = None
        self.clientId = None
        self.recordUid = None
        self.recordKey = None
        self.folderUid = None
        self.folderKey = None
        self.data = None


class DeletePayload:
    def __init__(self):
        self.clientVersion = None
        self.clientId = None
        self.recordUids = None


class FileUploadPayload:

    def __init__(self):
        self.clientVersion = None
        self.clientId = None
        self.fileRecordUid = None
        self.fileRecordKey = None
        self.fileRecordData = None
        self.ownerRecordUid = None
        self.ownerRecordData = None
        self.linkKey = None
        self.fileSize = None


class UpdatePayload:

    def __init__(self):
        self.clientVersion = None
        self.clientId = None
        self.recordUid = None
        self.data = None
        self.revision = None


class EncryptedPayload:

    def __init__(self, encrypted_payload, signature):
        self.encrypted_payload = encrypted_payload
        self.signature = signature


class KSMHttpResponse:

    def __init__(self, status_code, data, http_response=None):
        self.status_code = status_code
        self.data = data
        self.http_response = http_response



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com

import base64
import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.serialization import load_der_private_key

from . import utils


class CryptoUtils:

    BS = 16

    pad_binary = lambda s: s + (
                (CryptoUtils.BS - len(s) % CryptoUtils.BS) * chr(CryptoUtils.BS - len(s) % CryptoUtils.BS)).encode()

    unpad_binary = lambda s: s[0:-s[-1]]

    unpad_char = lambda s: s[0:-ord(s[-1])]

    @staticmethod
    def bytes_to_int(b):
        return int.from_bytes(b, byteorder='big')

    @staticmethod
    def url_safe_str_to_bytes(s):
        b = base64.urlsafe_b64decode(s + '==')
        return b

    @staticmethod
    def generate_random_bytes(length):
        return os.urandom(length)

    @staticmethod
    def generate_encryption_key_bytes():
        """Generates 32 bit transmission key"""
        return CryptoUtils.generate_random_bytes(32)

    @staticmethod
    def bytes_to_url_safe_str(b):
        return base64.urlsafe_b64encode(b).decode().rstrip('=')

    @staticmethod
    def url_safe_str_to_int(s):
        b = CryptoUtils.url_safe_str_to_bytes(s)
        return CryptoUtils.bytes_to_int(b)

    @staticmethod
    def generate_ecc_keys():

        encryption_key_bytes = CryptoUtils.generate_encryption_key_bytes()
        private_key_str = CryptoUtils.bytes_to_url_safe_str(encryption_key_bytes)
        encryption_key_int = CryptoUtils.url_safe_str_to_int(private_key_str)
        private_key = ec.derive_private_key(encryption_key_int, ec.SECP256R1(), default_backend())

        return private_key

    @staticmethod
    def public_key_ecc(private_key):
        pub_key = private_key.public_key()
        pub_key_bytes = pub_key.public_bytes(serialization.Encoding.X962, serialization.PublicFormat.UncompressedPoint)
        return pub_key_bytes

    @staticmethod
    def generate_private_key_ecc():
        encryption_key_bytes = utils.generate_random_bytes(32)
        private_key_str = utils.bytes_to_base64(encryption_key_bytes)

        encryption_key_int = utils.url_safe_str_to_int(private_key_str)

        private_key = ec.derive_private_key(encryption_key_int, ec.SECP256R1(), default_backend())

        return private_key

    @staticmethod
    def generate_private_key_der():

        private_key = CryptoUtils.generate_private_key_ecc()

        # export to DER

        private_key_der = private_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

        return private_key_der

    @staticmethod
    def generate_new_ecc_key():
        curve = ec.SECP256R1()
        ephemeral_key = ec.generate_private_key(curve, default_backend())
        return ephemeral_key

    @staticmethod
    def encrypt_aes(data, key, iv=None):
        aesgcm = AESGCM(key)
        iv = iv or os.urandom(12)
        enc = aesgcm.encrypt(iv, data, None)
        return iv + enc

    @staticmethod
    def public_encrypt(data: bytes, server_public_raw_key_bytes: bytes, idz: bytes = None):

        curve = ec.SECP256R1()

        ephemeral_public_key = ec.EllipticCurvePublicKey.from_encoded_point(curve, server_public_raw_key_bytes)

        # server_public_key = load_pem_public_key(server_public_key_bytes)
        ephemeral_key2 = CryptoUtils.generate_new_ecc_key()
        shared_key = ephemeral_key2.exchange(ec.ECDH(), ephemeral_public_key)

        if idz:
            shared_key = shared_key + idz

        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(shared_key)
        enc_key = digest.finalize()
        encrypted_data = CryptoUtils.encrypt_aes(data, enc_key)
        eph_public_key = ephemeral_key2.public_key().public_bytes(serialization.Encoding.X962,
                                                                  serialization.PublicFormat.UncompressedPoint)

        return eph_public_key + encrypted_data

    @staticmethod
    def hash_of_string(value):
        value_bytes = utils.base64_to_bytes(value)
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(value_bytes)
        return digest.finalize()

    @staticmethod
    def ecies_decrypt(server_public_key, ciphertext, priv_key_data, id=b''):

        try:
            # server_public_key = ciphertext[:65]
            encrypted_data = ciphertext
            curve = ec.SECP256R1()
            private_value = int.from_bytes(priv_key_data, byteorder='big', signed=False)
            ecc_private_key = ec.derive_private_key(private_value, curve, default_backend())

            # ecc_private_key = CommonHelperMethods.get_private_key_ecc(params)
            ephemeral_public_key = ec.EllipticCurvePublicKey.from_encoded_point(curve, server_public_key)
            shared_key = ecc_private_key.exchange(ec.ECDH(), ephemeral_public_key)
            shared_key = shared_key + id if id else shared_key
            digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
            digest.update(shared_key)
            enc_key = digest.finalize()
            result = CryptoUtils.decrypt_aes(encrypted_data, enc_key)

        except Exception as e:
            raise e

        return result

    @staticmethod
    def decrypt_aes(data, key):
        aesgcm = AESGCM(key)
        return aesgcm.decrypt(data[:12], data[12:], None)

    @staticmethod
    def decrypt_record(data, secret_key):
        if isinstance(data, str):
            data = utils.base64_to_bytes(data)

        record = CryptoUtils.decrypt_aes(data, secret_key)
        record_json = utils.bytes_to_string(record)
        return record_json

    @staticmethod
    def decrypt_ec(ecc_private_key, encrypted_data_bag: bytes):
        curve = ec.SECP256R1()

        server_public_key = encrypted_data_bag[:65]
        encrypted_data = encrypted_data_bag[65:]

        ephemeral_public_key = ec.EllipticCurvePublicKey.from_encoded_point(curve, server_public_key)
        shared_key = ecc_private_key.exchange(ec.ECDH(), ephemeral_public_key)
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(shared_key)
        enc_key = digest.finalize()
        decrypted_data = CryptoUtils.decrypt_aes(encrypted_data, enc_key)
        return decrypted_data

    @staticmethod
    def der_base64_private_key_to_private_key(private_key_der_base64):

        if isinstance(private_key_der_base64, str):
            private_key_der_base64 = utils.base64_to_bytes(private_key_der_base64)

        return load_der_private_key(private_key_der_base64, password=None)

    @staticmethod
    def extract_public_key_bytes(private_key_der_base64):

        if isinstance(private_key_der_base64, str):
            private_key_der_base64 = utils.base64_to_bytes(private_key_der_base64)

        ec_private_key = CryptoUtils.der_base64_private_key_to_private_key(private_key_der_base64)
        pub_key = ec_private_key.public_key()
        pub_key_bytes = pub_key.public_bytes(serialization.Encoding.X962, serialization.PublicFormat.UncompressedPoint)
        return pub_key_bytes

    @staticmethod
    def sign(data, private_key):
        signature = private_key.sign(
            data,
            ec.ECDSA(hashes.SHA256())
        )
        return signature



# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ (R)
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
import json
import logging
import os
from urllib.parse import urlparse

from keeper_secrets_manager_core.configkeys import ConfigKeys
from keeper_secrets_manager_core.keeper_globals import keeper_servers, logger_name
from keeper_secrets_manager_core.storage import KeyValueStorage


def get_server(code_server, config_store: KeyValueStorage):

    env_server = os.getenv('KSM_HOSTNAME')

    if env_server:
        server_to_use = env_server
    elif config_store.get(ConfigKeys.KEY_HOSTNAME):
        server_to_use = config_store.get(ConfigKeys.KEY_HOSTNAME)
    elif code_server:
        server_to_use = code_server
    else:
        server_to_use = keeper_servers.get('US')

    if server_to_use in keeper_servers:
        # Server key was supplied
        server_to_return = keeper_servers.get(server_to_use)
    else:
        # Looks like an URL
        # Un-parsing URL to get only domain:
        if 'http' not in server_to_use:
            server_to_use = 'https://%s' % server_to_use

        server_to_return = urlparse(server_to_use).netloc

    logging.getLogger(logger_name).debug("Keeper hostname %s" % server_to_return)

    return server_to_return


def is_json(json_str):
    try:
        json.loads(json_str)
    except ValueError as e:
        return False

    return True


def obj_to_dict(obj):
    return json.loads(
        json.dumps(obj, default=lambda o: getattr(o, '__dict__', str(o)))
    )


def get_folder_key(folder_uid, secrets_and_folders):

    folders = secrets_and_folders.folders

    for f in folders:
        if f.uid == folder_uid:
            return f

    return None
